import datetime
import json
import tempfile
from typing import Optional, Any

from adminboundarymanager.models import AdminBoundarySettings, AdminBoundary
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.core.files.base import File
from django.http import JsonResponse, HttpResponse
from django.shortcuts import get_object_or_404, render
from django.template.defaultfilters import filesizeformat
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django_large_image import tilesource
from large_image.exceptions import TileSourceXYZRangeError
from rest_framework.renderers import JSONRenderer
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from shapely import wkb
from wagtail.admin.auth import (
    user_passes_test,
    user_has_any_page_permission,
    permission_denied
)
from wagtail.models import Site
from wagtail.snippets.permissions import get_permission_name
from wagtail_modeladmin.helpers import AdminURLHelper
from wagtailcache.cache import cache_page

from geomanager.errors import RasterFileNotFound, QueryParamRequired, GeostoreNotFound
from geomanager.forms import LayerRasterFileForm
from geomanager.models import (
    Dataset,
    RasterUpload,
    RasterFileLayer,
    LayerRasterFile, Geostore
)
from geomanager.models.core import GeomanagerSettings
from geomanager.serializers import RasterFileLayerSerializer
from geomanager.utils import UUIDEncoder
from geomanager.utils.raster_utils import (
    get_tile_source,
    read_raster_info,
    create_layer_raster_file,
    get_raster_pixel_data, get_geostore_data,
    check_raster_bounds_with_boundary,
    clip_netcdf, clip_geotiff,
    bounds_to_polygon
)

ALLOWED_RASTER_EXTENSIONS = ["tif", "tiff", "geotiff", "nc"]


@user_passes_test(user_has_any_page_permission)
def upload_raster_file(request, dataset_id=None, layer_id=None):
    permission = get_permission_name('change', Dataset)
    if not request.user.has_perm(permission):
        return permission_denied(request)

    site = Site.objects.get(is_default_site=True)
    layer_manager_settings = GeomanagerSettings.for_site(site)

    file_error_messages = {
        "invalid_file_extension": _(
            "Not a supported raster format. Supported formats: %(supported_formats)s."
        ) % {"supported_formats": ALLOWED_RASTER_EXTENSIONS},
        "file_too_large": _(
            "This file is too big (%(file_size)s). Maximum filesize %(max_filesize)s. "
            "You can adjust the Maximum upload file size under Geomanager settings"
        ),
        "file_too_large_unknown_size": _(
            "This file is too big. Maximum filesize %(max_filesize)s. "
            "You can adjust the Maximum upload file size under Geomanager settings."
        ) % {"max_filesize": filesizeformat(layer_manager_settings.max_upload_size_bytes)}}

    layer = None
    context = {}
    context.update(
        {
            "max_filesize": layer_manager_settings.max_upload_size_bytes,
            "allowed_extensions": ALLOWED_RASTER_EXTENSIONS,
            "error_max_file_size": file_error_messages["file_too_large_unknown_size"],
            "error_accepted_file_types": file_error_messages["invalid_file_extension"],
        }
    )

    dataset = get_object_or_404(Dataset, pk=dataset_id)

    admin_url_helper = AdminURLHelper(Dataset)
    dataset_list_url = admin_url_helper.get_action_url("index")
    layer_list_url = None
    layer_preview_url = None

    if layer_id:
        layer = get_object_or_404(RasterFileLayer, pk=layer_id)
        layer_admin_url_helper = AdminURLHelper(layer)
        layer_list_url = layer_admin_url_helper.get_action_url("index") + f"?dataset__id__exact={dataset.pk}"
        layer_preview_url = layer.preview_url

    context.update({
        "dataset": dataset,
        "layer": layer,
        "datasets_index_url": dataset_list_url,
        "layers_index_url": layer_list_url,
        "dataset_preview_url": dataset.preview_url,
        "layer_preview_url": layer_preview_url
    })

    gm_settings = GeomanagerSettings.for_request(request)

    crop_raster = gm_settings.crop_raster_to_country

    # Check if user is submitting
    if request.method == 'POST':
        files = request.FILES.getlist('files[]', None)
        upload_file = files[0]

        upload = RasterUpload.objects.create(file=upload_file, dataset=dataset)

        raster_metadata = read_raster_info(upload.file.path)

        abm_settings = AdminBoundarySettings.for_request(request)
        abm_extents = abm_settings.combined_countries_bounds
        abm_countries = abm_settings.countries_list

        if abm_extents and crop_raster and raster_metadata.get("bounds"):
            intersects_with_boundary, completely_inside_boundary = check_raster_bounds_with_boundary(
                raster_metadata.get("bounds"), abm_extents)

            # clipping raster to boundaries
            if not intersects_with_boundary:
                # no intersection. The raster is not within the set countries of interest
                return JsonResponse({
                    "success": False,
                    "error_message": _("You have set to clip uploaded raster files to the country(s) boundary but the"
                                       " uploaded file is not within the set countries. Make sure the uploaded file "
                                       "is within your country/countries or disable 'Crop raster to country' option "
                                       "in Geomanager settings")
                })

            elif not completely_inside_boundary:
                raster_driver = raster_metadata.get("driver")

                country_geoms = []
                for country in abm_countries:
                    code = country.get("code")
                    alpha3 = country.get("alpha3")

                    # query using code (2-letter code)
                    country_boundary = AdminBoundary.objects.filter(level=0, gid_0=code).first()

                    # query using alpha 3 (3-letter code)
                    if not country_boundary:
                        country_boundary = AdminBoundary.objects.filter(level=0, gid_0=alpha3).first()

                    if country_boundary:
                        shapely_geom = wkb.loads(country_boundary.geom.hex)
                        country_geoms.append(shapely_geom)
                    else:
                        # use bbox instead
                        bbox = country.get("bbox")
                        if bbox:
                            bounds_geom = bounds_to_polygon(bbox)
                            country_geoms.append(bounds_geom)

                union_polygon = country_geoms[0]
                for polygon in country_geoms[1:]:
                    union_polygon = union_polygon.union(polygon)

                if raster_driver == "netCDF":
                    clip_fn = clip_netcdf
                    suffix = ".nc"
                else:
                    clip_fn = clip_geotiff
                    suffix = ".tif"

                with tempfile.NamedTemporaryFile(suffix=suffix) as f:
                    clipped_raster = clip_fn(upload.file.path, union_polygon, f.name)
                    raster_metadata = read_raster_info(clipped_raster)

                    with open(clipped_raster, 'rb') as clipped_file:
                        file_obj = File(clipped_file, name=f.name)
                        upload.file.save(upload.file.name, file_obj, save=True)

        upload.raster_metadata = raster_metadata
        upload.save()

        query_set = RasterFileLayer.objects.filter(dataset=dataset)

        initial_data = {
            "layer": layer_id if layer_id else query_set.first()
        }

        form_kwargs = {}

        timestamps = raster_metadata.get("timestamps", None)

        if timestamps:
            form_kwargs.update({"nc_dates_choices": timestamps})
            initial_data.update({"nc_dates": timestamps})

        data_variables = raster_metadata.get("data_variables", None)

        layer_form = LayerRasterFileForm(queryset=query_set, initial=initial_data, **form_kwargs)
        layer_forms = []

        if data_variables and len(data_variables) > 0:
            for variable in data_variables:
                form_init_data = {**initial_data, "nc_data_variable": variable}
                l_form = LayerRasterFileForm(queryset=query_set, initial=form_init_data, **form_kwargs)
                layer_forms.append({"data_variable": variable, "form": l_form})

        ctx = {
            "dataset": dataset,
            "publish_action": reverse("geomanager_publish_raster", args=[upload.pk]),
            "delete_action": reverse("geomanager_delete_raster_upload", args=[upload.pk]),
        }

        response = {
            "success": True,
        }

        # we have more than one layer, render multiple forms
        if layer_forms:
            forms = []
            for form in layer_forms:
                ctx.update({**form})
                forms.append(render_to_string(
                    "geomanager/raster_file_edit_form.html",
                    ctx,
                    request=request,
                ))
            response.update({"forms": forms})
        else:
            ctx.update({"form": layer_form})
            form = render_to_string(
                "geomanager/raster_file_edit_form.html",
                ctx,
                request=request,
            )
            response.update({"form": form})

        return JsonResponse(response)

    return render(request, 'geomanager/raster_file_upload.html', context)


@user_passes_test(user_has_any_page_permission)
def publish_raster(request, upload_id):
    if request.method != 'POST':
        return JsonResponse({"message": _("Only POST allowed")})

    upload = RasterUpload.objects.get(pk=upload_id)

    if not upload:
        return JsonResponse({"message": _("upload not found")}, status=404)

    db_layer = get_object_or_404(RasterFileLayer, pk=request.POST.get('layer'))

    raster_metadata = upload.raster_metadata

    form_kwargs = {}
    timestamps = raster_metadata.get("timestamps", None)

    data = {
        "layer": db_layer,
        "time": request.POST.get('time'),
        "nc_data_variable": request.POST.get('nc_data_variable')
    }

    if request.POST.get("nc_dates"):
        data.update({"nc_dates": request.POST.getlist("nc_dates")})

    if timestamps:
        form_kwargs.update({"nc_dates_choices": timestamps})

    queryset = RasterFileLayer.objects.filter(dataset=upload.dataset)
    layer_form = LayerRasterFileForm(data=data, queryset=queryset, **form_kwargs)

    ctx = {
        "dataset": upload.dataset,
        "publish_action": reverse("geomanager_publish_raster", args=[upload.pk]),
        "delete_action": reverse("geomanager_delete_raster_upload", args=[upload.pk]),
        "form": layer_form
    }

    def get_response():
        return {
            "success": False,
            "form": render_to_string(
                "geomanager/raster_file_edit_form.html",
                ctx,
                request=request,
            ),
        }

    if layer_form.is_valid():
        layer = layer_form.cleaned_data['layer']
        time = layer_form.cleaned_data['time']
        nc_dates = layer_form.cleaned_data['nc_dates']
        nc_data_variable = layer_form.cleaned_data['nc_data_variable']

        if nc_dates:
            data_timestamps = raster_metadata.get("timestamps")

            for time_str in nc_dates:
                try:
                    index = data_timestamps.index(time_str)

                    d_time = datetime.datetime.fromisoformat(time_str)
                    d_time = timezone.make_aware(d_time, timezone.get_current_timezone())

                    exists = LayerRasterFile.objects.filter(layer=db_layer, time=d_time).exists()

                    if exists:
                        error_message = _("File with date %(time_str)s already exists for layer %(db_layer)s") % {
                            "time_str": time_str, "db_layer": db_layer}
                        layer_form.add_error("nc_dates", error_message)
                        return JsonResponse(get_response())

                    create_layer_raster_file(layer, upload, time=d_time, band_index=str(index),
                                             data_variable=nc_data_variable)
                except Exception as e:
                    layer_form.add_error(None, _("Error occurred. Try again"))
                    return JsonResponse(get_response())
            # cleanup upload
            upload.delete()
            return JsonResponse({"success": True, })
        elif nc_data_variable:
            exists = LayerRasterFile.objects.filter(layer=db_layer, time=time).exists()

            if exists:
                error_message = _("File with date %(time)s already exists for layer %(db_layer)s") % {
                    "time": time.isoformat(), "db_layer": db_layer}
                layer_form.add_error("time", error_message)
                return JsonResponse(get_response())

            create_layer_raster_file(layer, upload, time, data_variable=nc_data_variable)
            # cleanup upload
            upload.delete()
            return JsonResponse({"success": True, })
        else:
            exists = LayerRasterFile.objects.filter(layer=db_layer, time=time).exists()

            if exists:
                error_message = _("File with date %(time)s already exists for selected layer") % {
                    "time": time.isoformat()}
                layer_form.add_error("time", error_message)
                return JsonResponse(get_response())

            create_layer_raster_file(layer, upload, time)
        # cleanup upload
        upload.delete()
        return JsonResponse(
            {
                "success": True,
            }
        )
    else:
        return JsonResponse(get_response())


@user_passes_test(user_has_any_page_permission)
def delete_raster_upload(request, upload_id):
    if request.method != 'POST':
        return JsonResponse({"message": _("Only POST allowed")})

    upload = RasterUpload.objects.filter(pk=upload_id)

    if upload.exists():
        upload.first().delete()
    else:
        return JsonResponse({"success": True})

    return JsonResponse({"success": True, "layer_raster_file_id": upload_id, })


@user_passes_test(user_has_any_page_permission)
def preview_raster_layers(request, dataset_id, layer_id=None):
    dataset = get_object_or_404(Dataset, pk=dataset_id)

    base_absolute_url = request.scheme + '://' + request.get_host()

    dataset_admin_helper = AdminURLHelper(Dataset)
    dataset_list_url = dataset_admin_helper.get_action_url("index")

    image_file_layer_admin_helper = AdminURLHelper(RasterFileLayer)
    image_file_layer_list_url = image_file_layer_admin_helper.get_action_url("index")

    dataset_layers = RasterFileLayerSerializer(dataset.raster_file_layers, many=True, context={"request": request}).data

    selected_layer = None

    if layer_id:
        selected_layer = dataset.raster_file_layers.get(pk=layer_id)

    context = {
        "dataset": dataset,
        "dataset_layers": json.dumps(dataset_layers, cls=UUIDEncoder),
        "selected_layer": selected_layer,
        "datasets_index_url": dataset_list_url,
        "image_file_layer_list_url": image_file_layer_list_url,
        "file_raster_list_url": request.build_absolute_uri(reverse("file-raster-list")),
        "large_image_color_maps_url": request.build_absolute_uri(reverse("large-image-colormaps")),
        "file_raster_metadata_url": request.build_absolute_uri(reverse("file-raster-metadata", args=("0",))),
        "base_absolute_url": base_absolute_url,
    }
    return render(request, 'geomanager/raster_file_layer_preview.html', context)


class RasterDataMixin:
    def get_single_raster_file(self, request: Request, layer_id) -> LayerRasterFile:
        time = self.get_query_param(request, "time")

        if not time:
            raise QueryParamRequired(_("time param required"))

        raster_file = LayerRasterFile.objects.filter(layer=layer_id, time=time)

        if raster_file.exists():
            return raster_file.first()

        error_message = _("File not found matching 'layer': %(layer_id)s and 'time': %(time)s") % {"layer_id": layer_id,
                                                                                                   "time": time}
        raise RasterFileNotFound(error_message)

    def get_multiple_raster_files(self, request: Request, layer_id) -> [LayerRasterFile]:
        time_from = self.get_query_param(request, "time_from")
        time_to = self.get_query_param(request, "time_to")

        if not time_from and not time_to:
            raise QueryParamRequired(_("time_from or time_to param required"))

        time_filter = {}

        if time_to and time_from:
            time_filter.update({"time__range": [time_from, time_to]})
        elif time_from:
            time_filter.update({"time__gte": time_from})
        else:
            time_filter.update({"time__lte": time_to})

        raster_files = LayerRasterFile.objects.filter(layer=layer_id, **time_filter)

        return raster_files

    def get_coords(self, request: Request):
        x_coord = self.get_query_param(request, "x")
        y_coord = self.get_query_param(request, "y")

        if not x_coord:
            raise QueryParamRequired(_("x param required"))

        if not y_coord:
            return QueryParamRequired(_("y param required"))

        x_coord = float(x_coord)
        y_coord = float(y_coord)

        return x_coord, y_coord

    def get_pixel_data(self, request: Request, layer_id):
        raster_file = self.get_single_raster_file(request, layer_id)
        x_coord, y_coord = self.get_coords(request)

        pixel_data = get_raster_pixel_data(raster_file.file, x_coord, y_coord)

        return {"date": raster_file.time, "value": pixel_data}

    def get_geostore(self, request: Request):
        geostore_id = self.get_query_param(request, "geostore_id")

        if not geostore_id:
            raise QueryParamRequired(_("geostore_id param required"))

        try:
            geostore = Geostore.objects.get(pk=geostore_id)
        except ObjectDoesNotExist:
            error_message = _("Geostore with id %(geostore_id)s does not exist") % {"geostore_id": geostore_id}
            raise GeostoreNotFound(error_message)
        return geostore

    def get_raster_geostore_data(self, raster_file, geostore, value_type):
        geostore_data = get_geostore_data(raster_file.file, geostore)

    def get_query_param(self, request: Request, key: str, default: Optional[Any] = '') -> str:
        return request.query_params.get(key, str(default))


@method_decorator(cache_page, name='get')
class RasterTileView(RasterDataMixin, APIView):
    # TODO: Validate style query param thoroughly. If not validated, the whole app just exits without warning.
    # TODO: Cache getting layer style. We should not be querying the database each time for style
    def get(self, request, layer_id, z, x, y):
        try:
            raster_file = self.get_single_raster_file(request, layer_id)
        except QueryParamRequired as e:
            return HttpResponse(e.message, status=400)
        except RasterFileNotFound as e:
            return HttpResponse(e, status=404)

        fmt = self.get_query_param(request, "format", "png")
        projection = self.get_query_param(request, "projection", "EPSG:3857")
        style = self.get_query_param(request, "style")
        geostore_id = self.get_query_param(request, "geostore_id")

        if style:
            # explict request to use layer defined style. Mostly used for admin previews
            if style == "layer-style":
                layer_style = raster_file.layer.style
                if layer_style:
                    style = layer_style.get_style_as_json()
                else:
                    style = None
            else:
                # try validating style
                # TODO: do more thorough validation
                try:
                    style = json.loads(style)
                except Exception:
                    style = None
        else:
            layer_style = raster_file.layer.style
            if layer_style:
                style = layer_style.get_style_as_json()

        encoding = tilesource.format_to_encoding(fmt, pil_safe=True)

        options = {
            "encoding": encoding,
            "projection": projection,
            "style": style,
            "geostore_id": geostore_id
        }

        source = get_tile_source(path=raster_file.file, options=options)

        try:
            tile_binary = source.getTile(int(x), int(y), int(z))
        except TileSourceXYZRangeError as e:
            raise ValidationError(e)

        mime_type = source.getTileMimeType()

        return HttpResponse(tile_binary, content_type=mime_type)


@method_decorator(cache_page, name='get')
class RasterDataPixelView(RasterDataMixin, APIView):
    renderer_classes = [JSONRenderer]

    def get(self, request, layer_id):
        try:
            pixel_data = self.get_pixel_data(request, layer_id)
        except QueryParamRequired as e:
            return JsonResponse(e.serialize, status=400)
        except RasterFileNotFound as e:
            return JsonResponse({"message": "Raster file not found"}, status=404)

        return Response(pixel_data)


@method_decorator(cache_page, name='get')
class RasterDataPixelTimeseriesView(RasterDataMixin, APIView):
    renderer_classes = [JSONRenderer]

    def get(self, request, layer_id):
        try:
            raster_files = self.get_multiple_raster_files(request, layer_id)
            x_coord, y_coord = self.get_coords(request)
        except QueryParamRequired as e:
            return JsonResponse(e.serialize, status=400)
        except RasterFileNotFound as e:
            return JsonResponse({"message": e}, status=404)

        timeseries_data = []

        for raster_file in raster_files:
            pixel_data = get_raster_pixel_data(raster_file.file, x_coord, y_coord)
            timeseries_data.append({"date": raster_file.time, "value": pixel_data})

        return Response(timeseries_data)


@method_decorator(cache_page, name='get')
class RasterDataGeostoreView(RasterDataMixin, APIView):
    renderer_classes = [JSONRenderer]

    def get(self, request, layer_id):
        try:
            value_type = self.get_query_param(request, "value_type")
            geostore = self.get_geostore(request)
            raster_file = self.get_single_raster_file(request, layer_id)
            data = get_geostore_data(raster_file.file, geostore, value_type)
        except QueryParamRequired as e:
            return JsonResponse(e.serialize, status=400)
        except (RasterFileNotFound, GeostoreNotFound) as e:
            return JsonResponse({"message": e}, status=404)
        return Response(data)


@method_decorator(cache_page, name='get')
class RasterDataGeostoreTimeseriesView(RasterDataMixin, APIView):
    renderer_classes = [JSONRenderer]

    def get(self, request, layer_id):
        try:
            value_type = self.get_query_param(request, "value_type")
            raster_files = self.get_multiple_raster_files(request, layer_id)
            geostore = self.get_geostore(request)
        except QueryParamRequired as e:
            return JsonResponse(e.serialize, status=400)
        except (RasterFileNotFound, GeostoreNotFound) as e:
            return JsonResponse({"message": e}, status=404)

        timeseries_data = []

        for raster_file in raster_files:
            data = get_geostore_data(raster_file.file, geostore, value_type)

            if value_type and data.get(value_type):
                data = data[value_type]
            else:
                data = data.get("mean")

            timeseries_data.append({"date": raster_file.time, "value": data})

        return Response(timeseries_data)


def raster_file_as_tile_json(request, layer_id):
    layer = get_object_or_404(RasterFileLayer, pk=layer_id)
    tile_json = layer.get_tile_json(request)
    return JsonResponse(tile_json)



import math
import pathlib
import tempfile

import numpy as np
import pandas as pd
import rasterio as rio
import xarray as xr
from django.core.files import File
from django.forms import FileField
from django_large_image import tilesource
from django_large_image.utilities import (
    field_file_to_local_path,
    get_cache_dir,
    get_file_lock,
    get_file_safe_path
)
from large_image.exceptions import TileSourceError
from rasterio import CRS
from rasterio.mask import mask
from rest_framework.exceptions import APIException
from rio_cogeo.cogeo import cog_translate
from rio_cogeo.profiles import cog_profiles
from shapely import wkb, Polygon

from geomanager.errors import UnsupportedRasterFormat
from geomanager.models import LayerRasterFile, Geostore


def get_tile_source(path, options=None):
    if options is None:
        options = {}

    encoding = options.get("encoding")
    style = options.get("style")
    projection = options.get("projection")
    geostore_id = options.get("geostore_id")

    kwargs = {}

    if encoding:
        kwargs["encoding"] = encoding
    if style:
        kwargs["style"] = style
    if projection:
        kwargs["projection"] = projection

    geostore = None

    if geostore_id:
        try:
            geostore = Geostore.objects.get(pk=geostore_id)
        except Exception:
            pass

    if geostore:
        file_path = field_file_to_local_path_for_geostore(path, geostore)
    else:
        file_path = field_file_to_local_path(path)
    try:
        return tilesource.get_tilesource_from_path(file_path, source=None, **kwargs)
    except (TileSourceError, Exception) as e:
        # Raise 500 server error if tile source failed to open
        raise APIException(str(e))


def get_raster_pixel_data(file: FileField, x_coord: float, y_coord: float):
    source = get_tile_source(path=file)

    # get raster affine transform
    metadata = source.getInternalMetadata()
    affine = metadata.get("Affine")

    # get corresponding row col
    row_col = rio.transform.rowcol(affine, xs=x_coord, ys=y_coord)
    pixel_data = source.getPixel(region={'left': abs(row_col[1]), 'top': abs(row_col[0])})

    if pixel_data:
        return pixel_data.get("bands", {}).get(1)

    return None


def get_geostore_data(file: FileField, geostore, value_type=None):
    data = {}

    try:
        file_path = field_file_to_local_path_for_geostore(file, geostore)

        with rio.open(str(file_path), 'r') as src:
            band_array = src.read(1)
            values = band_array[band_array != src.nodata]

        if value_type:
            if value_type == "mean":
                data.update({"mean": values.mean()})
            elif value_type == "sum":
                data.update({"sum": values.sum()})
            elif value_type == "minmax":
                data.update({
                    "min": values.min(),
                    "max": values.max()
                })
            elif value_type == "minmeanmax":
                data.update({
                    "min": values.min(),
                    "max": values.max(),
                    "mean": values.mean()
                })
            else:
                data.update({"mean": values.mean()})
        else:
            data.update({"mean": values.mean()})
    except Exception:
        pass

    return data


def get_no_data_val(val):
    if math.isnan(val):
        return None
    return val


def read_raster_info(file_path):
    raster = rio.open(file_path)

    no_data_vals = raster.nodatavals
    no_data_vals = [get_no_data_val(val) for val in no_data_vals]

    # get basic raster info using rasterio
    raster_info = {
        "crs": raster.crs.to_dict() if raster.crs else None,
        "bounds": raster.bounds,
        "width": raster.width,
        "height": raster.height,
        "bands_count": raster.count,
        "driver": raster.driver,
        "nodatavals": tuple(no_data_vals)
    }

    # close raster
    raster.close()

    # get netcdf info
    if raster_info["driver"] == "netCDF":

        ds = xr.open_dataset(file_path, decode_times=True)

        if isinstance(ds, xr.DataArray):
            ds = ds.to_dataset()

        skip_vars = ["nbnds", "time_bnds", "spatial_ref"]
        data_vars = list(ds.data_vars.keys())
        data_vars = [var for var in data_vars if var not in skip_vars]

        raster_info.update({"data_variables": data_vars})
        raster_info.update({"dimensions": list(ds.dims)})

        # get timestamps
        if "time" in ds.dims:
            timestamps = [pd.to_datetime(ts).isoformat() for ts in ds.time.data]

            raster_info.update({"timestamps": timestamps})

        # close dataset
        ds.close()

    return raster_info


def convert_upload_to_geotiff(upload, out_file_path, band_index=None, data_variable=None):
    metadata = upload.raster_metadata

    driver = metadata.get("driver")

    crs = metadata.get("crs")
    timestamps = metadata.get("timestamps", None)

    # handle netcdf
    if driver == "netCDF":
        rds = xr.open_dataset(upload.file.path, engine="rasterio")

        try:  # index must start from 0
            if data_variable:
                rds = rds[data_variable]

            if timestamps and band_index:
                rds = rds.isel(time=int(band_index))

            # write crs if not available
            if not rds.rio.crs:
                epsg = "epsg:4326"
                if crs and isinstance(crs, dict) and crs.get("init"):
                    epsg = crs.get("init")
                rds = rds.rio.write_crs(epsg)

            # drop grid_mapping attr. somehow it causes errors when saving
            if rds.rio.crs and rds.attrs.get("grid_mapping"):
                rds.attrs.pop("grid_mapping")

            # make sure no data value is not nan
            nodata_value = rds.encoding.get('nodata', rds.encoding.get('_FillValue'))
            if np.isnan(nodata_value):
                rds = rds.rio.write_nodata(-9999, encoded=True)

            netcdf_attrs = []
            for key in rds.attrs.keys():
                if key.startswith("NETCDF_"):
                    netcdf_attrs.append(key)

                # assign only one unit
                if key == "units" and isinstance(rds.attrs["units"], list):
                    rds.attrs["units"] = rds.attrs["units"][0]

            # delete 'NETCDF_*' attributes
            for nc_attr in netcdf_attrs:
                rds.attrs.pop(nc_attr)

            rds.rio.to_raster(out_file_path, driver="COG", compress="DEFLATE")
        except Exception as e:
            raise e
        finally:
            rds.close()

        return True

    # handle geotiff
    if driver == "GTiff":
        output_profile = cog_profiles.get("deflate")

        # write crs if not available
        if not crs:
            output_profile["crs"] = "epsg:4326"

        # save as COG
        cog_translate(
            upload.file.path,
            out_file_path,
            output_profile,
            indexes=band_index,
            in_memory=False
        )

        return True

    raise UnsupportedRasterFormat


def create_layer_raster_file(layer, upload, time, band_index=None, data_variable=None):
    with tempfile.NamedTemporaryFile(suffix=".tif") as f:
        convert_upload_to_geotiff(upload, f.name, band_index=band_index, data_variable=data_variable)
        with open(f.name, mode='rb') as file:
            file_content = File(file)
            raster = LayerRasterFile(layer=layer, time=time)
            file_name = f"{time.isoformat()}.tif"
            if data_variable:
                file_name = f"{data_variable}_{file_name}"
            raster.file.save(file_name, file_content)
            raster.save()


def clip_geotiff(geotiff_path, geom, out_file):
    data = rio.open(geotiff_path)
    out_img, out_transform = mask(data, shapes=[geom], crop=True)
    out_meta = data.meta.copy()
    out_meta.update({
        "driver": "GTiff",
        "height": out_img.shape[1],
        "width": out_img.shape[2],
        "transform": out_transform,
        "crs": CRS().from_epsg(code=4326),
    })

    with rio.open(out_file, "w", **out_meta) as dest:
        dest.write(out_img)

    return out_file


def clip_netcdf(nc_path, geom, out_file):
    rds = xr.open_dataset(nc_path, engine="rasterio")

    # write crs
    rds.rio.write_crs("epsg:4326", inplace=True)

    # clip
    rds = rds.rio.clip([geom], "epsg:4326", drop=True)

    # write clipped data to file
    rds.to_netcdf(out_file)

    rds.close()

    return out_file


def field_file_to_local_path_for_geostore(path, geostore):
    field_file_basename = pathlib.PurePath(path.name).name
    directory = get_cache_dir() / f"{type(path.instance).__name__}-{path.instance.pk}" / "geostore"
    dest_path = directory / f"{geostore.pk.hex}-{field_file_basename}"
    lock = get_file_lock(dest_path)
    safe = get_file_safe_path(dest_path)

    with lock.acquire():
        if not safe.exists():
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            # convert OGRGeometry to Shapely geometry for consistency in clipping
            shapely_geom = wkb.loads(geostore.geom.hex)
            clip_geotiff(path.file.name, shapely_geom, dest_path)

    return dest_path


def bounds_to_polygon(bounds):
    polygon_coords = ((bounds[0], bounds[1]), (bounds[2], bounds[1]), (bounds[2], bounds[3]), (bounds[0], bounds[3]))
    return Polygon(polygon_coords)


def check_raster_bounds_with_boundary(raster_bounds, boundary_bounds):
    boundary_poly = bounds_to_polygon(boundary_bounds)
    raster_poly = bounds_to_polygon(raster_bounds)

    intersects = boundary_poly.intersects(raster_poly)
    contains = boundary_poly.contains(raster_poly)

    return intersects, contains



from django.urls import include, path
from rest_framework.routers import SimpleRouter
from rest_framework_simplejwt.views import (
    TokenRefreshView, )
from wagtailcache.cache import nocache_page

from geomanager.viewsets.aoi import AoiViewSet
from .views import (
    RasterTileView,
    VectorTileView,
    map_view,
    RegisterView,
    ResetPasswordView,
    tile_gl,
    tile_json_gl,
    style_json_gl,
    get_mapviewer_config,
    GeoJSONPgTableView
)
from .views.auth import (
    EmailTokenObtainPairView,
    UserTokenVerifyView
)
from .views.profile import (
    get_geomanager_user_profile,
    create_or_update_geomanager_user_profile
)
from .views.raster_file import (
    RasterDataPixelView,
    RasterDataPixelTimeseriesView,
    RasterDataGeostoreView,
    RasterDataGeostoreTimeseriesView, raster_file_as_tile_json
)
from .views.stations import StationsTileView
from .viewsets import (
    RasterLayerRasterFileDetailViewSet,
    VectorTableFileDetailViewSet,
    DatasetViewSet,
    GeostoreViewSet,
    AdminBoundaryViewSet,
    MetadataViewSet
)

router = SimpleRouter(trailing_slash=True)

router.register(r'api/datasets', DatasetViewSet)
router.register(r'api/metadata', MetadataViewSet)

router.register(r'api/file-raster', RasterLayerRasterFileDetailViewSet, basename="file-raster")
router.register(r'api/vector-data', VectorTableFileDetailViewSet)

router.register(r'api/aoi', AoiViewSet)

urlpatterns = [
                  # MapViewer
                  path(r'mapviewer/', map_view, name="mapview"),
                  path(r'mapviewer/<str:location_type>/', map_view, name="mapview"),
                  path(r'mapviewer/<str:location_type>/<str:adm0>/', map_view, name="mapview"),
                  path(r'mapviewer/<str:location_type>/<str:adm0>/<str:adm1>/', map_view, name="mapview"),
                  path(r'mapviewer/<str:location_type>/<str:adm0>/<str:adm1>/', map_view, name="mapview"),
                  path(r'mapviewer/<str:location_type>/<str:adm0>/<str:adm1>/<str:adm2>/', map_view, name="mapview"),

                  # MapViewer configuration
                  path(r'api/mapviewer-config', get_mapviewer_config, name="mapview_config"),

                  # Authentication
                  path('api/auth/register/', RegisterView.as_view(), name='auth_register'),
                  path('api/auth/reset-password/', ResetPasswordView.as_view(), name='auth_password_reset'),
                  path('api/auth/token/', EmailTokenObtainPairView.as_view(), name='token_obtain_pair'),
                  path('api/auth/token/verify/', nocache_page(UserTokenVerifyView.as_view()), name='token_verify'),
                  path('api/auth/token/refresh/', nocache_page(TokenRefreshView.as_view()), name='token_refresh'),

                  # User Profile
                  path('api/geomanager-profile/<int:user_id>', get_geomanager_user_profile,
                       name='get_geomanager_user_profile'),
                  path('api/geomanager-profile/update/<str:user_id>', create_or_update_geomanager_user_profile,
                       name='update_geomanager_user_profile'),

                  # # User Areas of Interest
                  # path('api/aoi/<int:user_id>', get_user_aoi_list, name='get_user_aoi_list'),
                  # path('api/aoi/', create_aoi, name='create_user_aoi'),
                  # path('api/aoi/', create_aoi, name='create_user_aoi'),

                  # Country
                  path(r'api/country', AdminBoundaryViewSet.as_view({"get": "get"}), name="country_list"),
                  path(r'api/country/<str:gid_0>', AdminBoundaryViewSet.as_view({"get": "get_regions"}),
                       name="country_regions"),
                  path(r'api/country/<str:gid_0>/<str:gid_1>',
                       AdminBoundaryViewSet.as_view({"get": "get_sub_regions"}),
                       name="country_sub_regions"),

                  # Geostore
                  path(r'api/geostore/', GeostoreViewSet.as_view({"post": "post"}), name="geostore"),
                  path(r'api/geostore/<uuid:geostore_id>', GeostoreViewSet.as_view({"get": "get"}),
                       name="get_by_geostore"),
                  path(r'api/geostore/admin/<str:gid_0>', GeostoreViewSet.as_view({"get": "get_by_admin"}),
                       name="get_by_gid0"),
                  path(r'api/geostore/admin/<str:gid_0>/<str:gid_1>', GeostoreViewSet.as_view({"get": "get_by_admin"}),
                       name="get_by_gid1"),
                  path(r'api/geostore/admin/<str:gid_0>/<str:gid_1>/<str:gid_2>',
                       GeostoreViewSet.as_view({"get": "get_by_admin"}),
                       name="get_by_gid2"),

                  # Tile JSON
                  path(r'api/raster/<uuid:layer_id>/tiles.json', raster_file_as_tile_json,
                       name="raster_file_tile_json"),

                  # Tiles
                  path(r'api/raster-tiles/<uuid:layer_id>/<int:z>/<int:x>/<int:y>', RasterTileView.as_view(),
                       name="raster_tiles"),
                  path(r'api/vector-tiles/<int:z>/<int:x>/<int:y>', VectorTileView.as_view(), name="vector_tiles"),
                  path(r'api/station-tiles/<int:z>/<int:x>/<int:y>', StationsTileView.as_view(), name="station_tiles"),

                  # Data
                  path(r'api/raster-data/pixel/<uuid:layer_id>', RasterDataPixelView.as_view(),
                       name="raster_data_pixel"),
                  path(r'api/raster-data/pixel/timeseries/<uuid:layer_id>', RasterDataPixelTimeseriesView.as_view(),
                       name="raster_data_pixel_timeseries"),

                  path(r'api/raster-data/geostore/<uuid:layer_id>', RasterDataGeostoreView.as_view(),
                       name="raster_data_geostore"),
                  path(r'api/raster-data/geostore/timeseries/<uuid:layer_id>',
                       RasterDataGeostoreTimeseriesView.as_view(),
                       name="raster_data_geostore_timeseries"),

                  # FeatureServ
                  path(r'api/feature-serv/<str:table_name>.geojson', GeoJSONPgTableView.as_view(),
                       name="feature_serv"),

                  # Tiles GL
                  path(r'api/tile-gl/tile/<str:source_slug>/<int:z>/<int:x>/<int:y>.pbf', tile_gl, name="tile_gl"),
                  path(r'api/tile-gl/tile-json/<str:source_slug>.json', tile_json_gl, name="tile_json_gl"),
                  path(r'api/tile-gl/style/<str:source_slug>.json', style_json_gl, name="style_json_gl"),

                  # Additional, standalone URLs from django-large-image
                  path('', include('django_large_image.urls')),
                  path('', include('adminboundarymanager.urls')),
              ] + router.urls



from adminboundarymanager.models import AdminBoundarySettings
from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.urls import path, reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from wagtail import hooks
from wagtail.admin.menu import MenuItem
from wagtail_adminsortable.admin import SortableAdminMixin
from wagtail_modeladmin.helpers import AdminURLHelper
from wagtail_modeladmin.options import ModelAdminGroup, ModelAdmin, modeladmin_register
from wagtail_modeladmin.views import CreateView, EditView, IndexView

from .helpers import (DatasetButtonHelper, CategoryButtonHelper, FileLayerButtonHelper)
from .models import (
    SubCategory, GeomanagerSettings,
    Category,
    Dataset,
    Metadata,
    RasterFileLayer,
    MBTSource, WmsLayer, RasterStyle, LayerRasterFile,
    VectorFileLayer, PgVectorTable, RasterTileLayer, VectorTileLayer
)
from .views import (
    upload_raster_file,
    publish_raster,
    delete_raster_upload,
    preview_raster_layers,
    upload_vector_file,
    publish_vector,
    delete_vector_upload,
    preview_vector_layers,
    preview_wms_layers,
    load_stations,
    preview_stations
)
from .views.raster_tile import preview_raster_tile_layers
from .views.vector_tile import preview_vector_tile_layers


@hooks.register('register_admin_urls')
def urlconf_geomanager():
    return [
        path('upload-rasters/', upload_raster_file, name='geomanager_upload_rasters'),
        path('upload-rasters/<uuid:dataset_id>/', upload_raster_file, name='geomanager_dataset_upload_raster'),
        path('upload-rasters/<uuid:dataset_id>/<uuid:layer_id>/', upload_raster_file,
             name='geomanager_dataset_layer_upload_raster'),

        path('publish-rasters/<int:upload_id>/', publish_raster, name='geomanager_publish_raster'),
        path('delete-raster-upload/<int:upload_id>/', delete_raster_upload, name='geomanager_delete_raster_upload'),

        path('preview-raster-layers/<uuid:dataset_id>/', preview_raster_layers,
             name='geomanager_preview_raster_dataset'),
        path('preview-raster-layers/<uuid:dataset_id>/<uuid:layer_id>/', preview_raster_layers,
             name='geomanager_preview_raster_layer'),
        path('upload-vector/', upload_vector_file, name='geomanager_upload_vector'),
        path('upload-vector/<uuid:dataset_id>/', upload_vector_file, name='geomanager_dataset_upload_vector'),
        path('upload-vector/<uuid:dataset_id>/<uuid:layer_id>/', upload_vector_file,
             name='geomanager_dataset_layer_upload_vector'),

        path('publish-vector/<int:upload_id>/', publish_vector, name='geomanager_publish_vector'),
        path('delete-vector-upload/<int:upload_id>/', delete_vector_upload, name='geomanager_delete_vector_upload'),

        path('preview-vector-layers/<uuid:dataset_id>/', preview_vector_layers,
             name='geomanager_preview_vector_dataset'),
        path('preview-vector-layers/<uuid:dataset_id>/<uuid:layer_id>/', preview_vector_layers,
             name='geomanager_preview_vector_layer'),

        path('preview-wms-layers/<uuid:dataset_id>/', preview_wms_layers,
             name='geomanager_preview_wms_dataset'),
        path('preview-wms-layers/<uuid:dataset_id>/<uuid:layer_id>/', preview_wms_layers,
             name='geomanager_preview_wms_layer'),

        path('preview-vector-tile-layers/<uuid:dataset_id>/', preview_vector_tile_layers,
             name='geomanager_preview_vector_tile_dataset'),
        path('preview-vector-tile-layers/<uuid:dataset_id>/<uuid:layer_id>/', preview_vector_tile_layers,
             name='geomanager_preview_vector_tile_layer'),

        path('preview-raster-tile-layers/<uuid:dataset_id>/', preview_raster_tile_layers,
             name='geomanager_preview_raster_tile_dataset'),
        path('preview-raster-tile-layers/<uuid:dataset_id>/<uuid:layer_id>/', preview_raster_tile_layers,
             name='geomanager_preview_raster_tile_layer'),

        path('load-stations/', load_stations, name='geomanager_load_stations'),
        path('preview-stations/', preview_stations, name='geomanager_preview_stations'),
    ]


class ModelAdminCanHide(ModelAdmin):
    hidden = False


class CategoryModelAdmin(SortableAdminMixin, ModelAdminCanHide):
    model = Category
    menu_label = _("Data Categories")
    exclude_from_explorer = True
    button_helper_class = CategoryButtonHelper
    menu_icon = "layer-group"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.list_display = (list(self.list_display) or []) + ["create_dataset", 'view_datasets', ]
        self.create_dataset.__func__.short_description = _('Create Dataset')
        self.view_datasets.__func__.short_description = _('View Datasets')

    def create_dataset(self, obj):
        label = _("Create Dataset")
        button_html = f"""
            <a href="{obj.dataset_create_url()}" class="button button-small button--icon bicolor">
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-plus"></use>
                    </svg>
                </span>
              {label}
            </a>
        """
        return mark_safe(button_html)

    def view_datasets(self, obj):
        label = _("View Datasets")
        button_html = f"""
            <a href="{obj.datasets_list_url()}" class="button button-small button--icon bicolor button-secondary">
                <span class="icon-wrapper">
                    <svg class="icon icon-list-ul icon" aria-hidden="true">
                        <use href="#icon-list-ul"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)


class DatasetIndexView(IndexView):
    def get_context_data(self, **kwargs):
        context_data = super(DatasetIndexView, self).get_context_data(**kwargs)

        categories_admin_helper = AdminURLHelper(Category)
        url = categories_admin_helper.get_action_url("index")

        context_data.update({
            "custom_create_url": {
                "label": _("Create from categories"),
                "url": url
            }
        })

        return context_data


class DatasetCreateView(CreateView):
    def get_form(self):
        form = super().get_form()
        category_id = self.request.GET.get("category_id")
        if category_id:
            # form.fields["category"].widget.attrs.update({"disabled": "true"})
            form.fields["sub_category"].queryset = SubCategory.objects.filter(category=category_id)
            initial = {**form.initial}
            initial.update({"category": category_id})
            form.initial = initial
        return form


class DatasetModelAdmin(ModelAdminCanHide):
    model = Dataset
    exclude_from_explorer = True
    button_helper_class = DatasetButtonHelper
    list_display = ("__str__", "layer_type",)
    list_filter = ("category", "id",)
    index_template_name = "modeladmin/index_without_custom_create.html"
    menu_icon = "database"

    index_view_class = DatasetIndexView
    create_view_class = DatasetCreateView

    def __init__(self, parent=None):
        super().__init__(parent)
        self.list_display = (list(self.list_display) or []) + ['category_link', 'view_layers', 'upload_files']
        self.category_link.__func__.short_description = _('Category')
        self.view_layers.__func__.short_description = _('View Layers')
        self.upload_files.__func__.short_description = _('Upload Files')

    def category_link(self, obj):
        label = _("Edit Category")
        button_html = f"""
            <a href="{obj.category_url}">
                {label}
            </a>
        """
        return mark_safe(button_html)

    def view_layers(self, obj):
        label = _("Layers")
        button_html = f"""
            <a href="{obj.layers_list_url()}" class="button button-small button--icon bicolor button-secondary">
                <span class="icon-wrapper">
                    <svg class="icon icon-list-ol icon" aria-hidden="true">
                        <use href="#icon-list-ol"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)

    def preview_dataset(self, obj):
        if not obj.preview_url:
            return None

        if obj.layer_type == "vector_file":
            return None

        disabled = "" if obj.can_preview() else "disabled"
        label = _("Preview Dataset")
        button_html = f"""
            <a href="{obj.preview_url}" class="button button-small button--icon button-secondary {disabled}">
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-view"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)

    def upload_files(self, obj):
        if not obj.upload_url:
            return None

        disabled = "" if obj.has_layers() else "disabled"
        label = _("Upload Files")
        button_html = f"""
            <a href="{obj.upload_url}" class="button button-small bicolor button--icon {disabled}">
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-upload"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)


class LayerIndexView(IndexView):
    def get_context_data(self, **kwargs):
        context_data = super(LayerIndexView, self).get_context_data(**kwargs)

        dataset_admin_helper = AdminURLHelper(Dataset)
        url = dataset_admin_helper.get_action_url("index")

        context_data.update({
            "custom_create_url": {
                "label": _("Create from datasets"),
                "url": url
            }
        })

        return context_data


class RasterFileLayerCreateView(CreateView):
    def get_form(self):
        form = super().get_form()
        form.fields["dataset"].queryset = Dataset.objects.filter(layer_type="raster_file")

        dataset_id = self.request.GET.get("dataset_id")
        if dataset_id:
            initial = {**form.initial}
            initial.update({"dataset": dataset_id})
            form.initial = initial
        return form


class RasterFileLayerEditView(EditView):
    def get_form(self):
        form = super().get_form()
        form.fields["dataset"].queryset = Dataset.objects.filter(layer_type="raster_file")
        return form


class RasterFileLayerModelAdmin(ModelAdminCanHide):
    model = RasterFileLayer
    hidden = True
    exclude_from_explorer = True
    menu_label = _("File Layers")
    button_helper_class = FileLayerButtonHelper
    index_view_class = LayerIndexView
    create_view_class = RasterFileLayerCreateView
    edit_view_class = RasterFileLayerEditView
    list_display = ("title",)
    list_filter = ("dataset",)
    index_template_name = "modeladmin/index_without_custom_create.html"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.list_display = (list(self.list_display) or []) + ["dataset_link", "upload_files", "uploaded_files",
                                                               "preview_layer"]
        self.dataset_link.__func__.short_description = _("Dataset")
        self.uploaded_files.__func__.short_description = _("Uploaded Files")
        self.upload_files.__func__.short_description = _("Upload Raster Files")
        self.preview_layer.__func__.short_description = _("Preview on Map")

    def dataset_link(self, obj):
        button_html = f"""
        <a href="{obj.dataset.dataset_url()}">
        {obj.dataset.title}
        </a>
        """
        return mark_safe(button_html)

    def upload_files(self, obj):
        label = _("Upload Files")
        button_html = f"""
            <a href="{obj.upload_url}" class="button button-small bicolor button--icon">
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-upload"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)

    def uploaded_files(self, obj):
        label = _("View Uploaded Files")
        button_html = f"""
            <a href="{obj.get_uploads_list_url()}" class="button button-small button--icon bicolor button-secondary">
                <span class="icon-wrapper">
                    <svg class="icon icon-list-ol icon" aria-hidden="true">
                        <use href="#icon-list-ol"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)

    def preview_layer(self, obj):
        label = _("Preview Layer")
        button_html = f"""
            <a href="{obj.preview_url}" class="button button-small button--icon button-secondary">
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-view"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)


class MetadataModelAdmin(ModelAdminCanHide):
    model = Metadata
    exclude_from_explorer = True
    menu_icon = 'info-circle'


class ModelAdminGroupWithHiddenItems(ModelAdminGroup):
    def get_submenu_items(self):
        menu_items = []
        item_order = 1
        for model_admin in self.modeladmin_instances:
            if not model_admin.hidden:
                menu_items.append(model_admin.get_menu_item(order=item_order))
                item_order += 1
        return menu_items


class RasterStyleCreateView(CreateView):
    def get_form(self):
        form = super().get_form()

        layer_id = self.request.GET.get("layer_id")

        # add hidden layer_id field to form. We will use it later to update the layer style
        if layer_id:
            try:
                layer = RasterFileLayer.objects.get(pk=layer_id)
                form.fields["layer_id"] = forms.CharField(required=False, widget=forms.HiddenInput())
                form.initial.update({"layer_id": layer.pk})
            except ObjectDoesNotExist:
                pass

        return form

    def form_valid(self, form):
        response = super().form_valid(form)

        # check if we have layer_id in data
        layer_id = form.data.get("layer_id")

        if layer_id:
            try:
                # assign this layer the just created style
                layer = RasterFileLayer.objects.get(pk=layer_id)
                layer.style = self.instance
                layer.save()
            except ObjectDoesNotExist:
                pass

        return response


class RasterStyleModelAdmin(ModelAdminCanHide):
    model = RasterStyle
    exclude_from_explorer = True
    create_view_class = RasterStyleCreateView
    list_display = ("__str__", "min", "max")
    form_view_extra_js = ["geomanager/js/raster_style_extra.js"]
    menu_icon = "palette"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.list_display = (list(self.list_display) or []) + ["preview"]
        self.preview.__func__.short_description = _("Color Preview")

    def preview(self, obj):
        if obj.use_custom_colors:
            return None
        color_list = [f"<li style='background-color:{color};height:20px;flex:1;'><li/>" for color in
                      obj.palette.split(",")]
        html = f"""
            <ul style='display:flex;width:200px;box-shadow: 0 1px 6px rgba(0, 0, 0, 0.12), 0 1px 4px rgba(0, 0, 0, 0.12);'>
                {''.join(color_list)}
            </ul>
        """
        return mark_safe(html)


class WmsLayerCreateView(CreateView):
    def get_form(self):
        form = super().get_form()
        form.fields["dataset"].queryset = Dataset.objects.filter(layer_type="wms")

        dataset_id = self.request.GET.get("dataset_id")
        if dataset_id:
            initial = {**form.initial}
            initial.update({"dataset": dataset_id})
            form.initial = initial
        return form


class WmsLayerModelAdmin(ModelAdminCanHide):
    model = WmsLayer
    exclude_from_explorer = True
    create_view_class = WmsLayerCreateView
    hidden = True
    index_template_name = "modeladmin/index_without_custom_create.html"
    index_view_class = LayerIndexView

    def __init__(self, parent=None):
        super().__init__(parent)
        self.list_display = (list(self.list_display) or []) + ['dataset_link', 'preview_layer']
        self.dataset_link.__func__.short_description = _('Dataset')
        self.preview_layer.__func__.short_description = _('Preview on Map')

    def dataset_link(self, obj):
        button_html = f"""
            <a href="{obj.dataset.dataset_url()}">
                {obj.dataset.title}
            </a>
        """
        return mark_safe(button_html)

    def preview_layer(self, obj):
        label = _("Preview Layer")
        button_html = f"""
            <a href="{obj.preview_url}" class="button button-small button--icon button-secondary">
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-view"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)


class RasterTileLayerCreateView(CreateView):
    def get_form(self):
        form = super().get_form()
        form.fields["dataset"].queryset = Dataset.objects.filter(layer_type="raster_tile")

        dataset_id = self.request.GET.get("dataset_id")
        if dataset_id:
            initial = {**form.initial}
            initial.update({"dataset": dataset_id})
            form.initial = initial
        return form


class RasterTileLayerModelAdmin(ModelAdminCanHide):
    model = RasterTileLayer
    exclude_from_explorer = True
    create_view_class = RasterTileLayerCreateView
    hidden = True
    index_template_name = "modeladmin/index_without_custom_create.html"
    index_view_class = LayerIndexView
    form_view_extra_js = ["geomanager/js/raster_tile_layer_form.js"]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.list_display = (list(self.list_display) or []) + ['dataset_link', 'preview_layer']
        self.dataset_link.__func__.short_description = _('Dataset')
        self.preview_layer.__func__.short_description = _('Preview on Map')

    def dataset_link(self, obj):
        button_html = f"""
            <a href="{obj.dataset.dataset_url()}">
                {obj.dataset.title}
            </a>
        """
        return mark_safe(button_html)

    def preview_layer(self, obj):
        label = _("Preview Layer")
        button_html = f"""
            <a href="{obj.preview_url}" class="button button-small button--icon button-secondary">
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-view"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)


class VectorTileLayerCreateView(CreateView):
    def get_form(self):
        form = super().get_form()
        form.fields["dataset"].queryset = Dataset.objects.filter(layer_type="vector_tile")

        dataset_id = self.request.GET.get("dataset_id")
        if dataset_id:
            initial = {**form.initial}
            initial.update({"dataset": dataset_id})
            form.initial = initial
        return form


class VectorTileLayerModelAdmin(ModelAdminCanHide):
    model = VectorTileLayer
    exclude_from_explorer = True
    create_view_class = VectorTileLayerCreateView
    hidden = True
    index_template_name = "modeladmin/index_without_custom_create.html"
    index_view_class = LayerIndexView

    def __init__(self, parent=None):
        super().__init__(parent)
        self.list_display = (list(self.list_display) or []) + ['dataset_link', 'preview_layer']
        self.dataset_link.__func__.short_description = _('Dataset')
        self.preview_layer.__func__.short_description = _('Preview on Map')

    def dataset_link(self, obj):
        button_html = f"""
            <a href="{obj.dataset.dataset_url()}">
                {obj.dataset.title}
            </a>
        """
        return mark_safe(button_html)

    def preview_layer(self, obj):
        label = _("Preview Layer")
        button_html = f"""
            <a href="{obj.preview_url}" class="button button-small button--icon button-secondary">
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-view"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)


class VectorFileLayerCreateView(CreateView):
    def get_form(self):
        form = super().get_form()
        form.fields["dataset"].queryset = Dataset.objects.filter(layer_type="vector_file")

        dataset_id = self.request.GET.get("dataset_id")
        if dataset_id:
            initial = {**form.initial}
            initial.update({"dataset": dataset_id})
            form.initial = initial
        return form


class VectorFileLayerEditView(EditView):
    def get_form(self):
        form = super().get_form()
        form.fields["dataset"].queryset = Dataset.objects.filter(layer_type="vector_file")
        return form


class VectorFileLayerModelAdmin(ModelAdminCanHide):
    model = VectorFileLayer
    hidden = True
    exclude_from_explorer = True
    menu_label = _("Vector Layers")
    index_view_class = LayerIndexView
    create_view_class = VectorFileLayerCreateView
    edit_view_class = VectorFileLayerEditView
    list_display = ("title",)
    list_filter = ("dataset",)
    index_template_name = "modeladmin/index_without_custom_create.html"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.list_display = (list(self.list_display) or []) + ['dataset_link', "uploaded_files", "upload_files",
                                                               'preview_layer']
        self.dataset_link.__func__.short_description = _('Dataset')
        self.uploaded_files.__func__.short_description = _("View Uploaded Files")
        self.upload_files.__func__.short_description = _('Upload Vector Files')
        self.preview_layer.__func__.short_description = _('Preview on Map')

    def dataset_link(self, obj):
        button_html = f"""
            <a href="{obj.dataset.dataset_url()}">
            {obj.dataset.title}
            </a>
        """
        return mark_safe(button_html)

    def upload_files(self, obj):
        disabled = "" if not obj.has_data_table else "disabled"

        label = _("Upload Files")
        button_html = f"""
            <a href="{obj.upload_url}" class="button button-small bicolor button--icon" {disabled}>
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-upload"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)

    def preview_layer(self, obj):
        label = _("Preview Layer")
        button_html = f"""
            <a href="{obj.preview_url}" class="button button-small button--icon button-secondary">
                <span class="icon-wrapper">
                    <svg class="icon icon-plus icon" aria-hidden="true">
                        <use href="#icon-view"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)

    def uploaded_files(self, obj):
        label = _("Uploaded Files")
        button_html = f"""
            <a href="{obj.get_uploads_list_url()}" class="button button-small button--icon bicolor button-secondary">
                <span class="icon-wrapper">
                    <svg class="icon icon-list-ol icon" aria-hidden="true">
                        <use href="#icon-list-ol"></use>
                    </svg>
                </span>
                {label}
            </a>
        """
        return mark_safe(button_html)


class MBTSourceModelAdmin(ModelAdminCanHide):
    model = MBTSource
    menu_label = _("Basemap Sources")
    menu_icon = "globe"
    form_view_extra_js = ["geomanager/js/mbt_source_extra.js"]


class LayerRasterFileModelAdmin(ModelAdminCanHide):
    model = LayerRasterFile
    exclude_from_explorer = True
    hidden = True
    list_display = ("__str__", "layer", "time",)
    list_filter = ("layer",)
    index_template_name = "modeladmin/index_without_custom_create.html"


class PgVectorTableModelAdmin(ModelAdminCanHide):
    model = PgVectorTable
    hidden = True
    list_display = ("__str__", "table_name",)
    list_filter = ("layer",)
    index_template_name = "modeladmin/index_without_custom_create.html"
    inspect_view_enabled = True


class GeoManagerAdminGroup(ModelAdminGroupWithHiddenItems):
    menu_label = _('Geo Manager')
    menu_icon = 'layer-group'
    menu_order = 700
    items = (
        CategoryModelAdmin,
        DatasetModelAdmin,
        MetadataModelAdmin,
        RasterFileLayerModelAdmin,
        RasterStyleModelAdmin,
        VectorFileLayerModelAdmin,
        WmsLayerModelAdmin,
        RasterTileLayerModelAdmin,
        VectorTileLayerModelAdmin,
        MBTSourceModelAdmin,
        LayerRasterFileModelAdmin,
        PgVectorTableModelAdmin
    )

    def get_submenu_items(self):
        menu_items = super().get_submenu_items()

        try:
            settings_url = reverse(
                "wagtailsettings:edit",
                args=[AdminBoundarySettings._meta.app_label, AdminBoundarySettings._meta.model_name, ],
            )
            abm_settings_menu = MenuItem(label=_("Boundary Settings"), url=settings_url, icon_name="cog")
            menu_items.append(abm_settings_menu)
        except Exception:
            pass

        boundary_loader = MenuItem(label=_("Boundary Data"), url=reverse("adminboundarymanager_preview_boundary"),
                                   icon_name="map")
        menu_items.append(boundary_loader)

        stations_data = MenuItem(label=_("Stations Data"), url=reverse("geomanager_preview_stations"),
                                 icon_name="map")
        menu_items.append(stations_data)

        try:
            settings_url = reverse(
                "wagtailsettings:edit",
                args=[GeomanagerSettings._meta.app_label, GeomanagerSettings._meta.model_name, ],
            )
            gm_settings_menu = MenuItem(label=_("Settings"), url=settings_url, icon_name="cog")
            menu_items.append(gm_settings_menu)
        except Exception:
            pass

        return menu_items


modeladmin_register(GeoManagerAdminGroup)


@hooks.register("register_icons")
def register_icons(icons):
    return icons + [
        'wagtailfontawesomesvg/solid/palette.svg',
        'wagtailfontawesomesvg/solid/database.svg',
        'wagtailfontawesomesvg/solid/layer-group.svg',
        'wagtailfontawesomesvg/solid/globe.svg',
        'wagtailfontawesomesvg/solid/map.svg',
    ]


@hooks.register('construct_settings_menu')
def hide_settings_menu_item(request, menu_items):
    hidden_settings = ["admin-boundary-settings", "geomanager-settings"]
    menu_items[:] = [item for item in menu_items if item.name not in hidden_settings]



from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from rest_framework import serializers
from rest_framework.validators import UniqueValidator
from rest_framework_simplejwt.serializers import TokenObtainPairSerializer


class EmailTokenObtainSerializer(TokenObtainPairSerializer):
    def validate(self, attrs):
        raw_username = attrs["username"]
        user = get_user_model().objects.filter(email=raw_username)
        if user:
            user = user.first()
            attrs['username'] = user.username

        data = super().validate(attrs)
        return data


class RegisterSerializer(serializers.ModelSerializer):
    username = serializers.EmailField(
        required=True,
        validators=[
            UniqueValidator(queryset=get_user_model().objects.all(), message="User with this email already exists")]
    )

    class Meta:
        model = get_user_model()
        fields = ('username',)


class ResetPasswordSerializer(serializers.ModelSerializer):
    username = serializers.EmailField(required=True)

    class Meta:
        model = get_user_model()
        fields = ('username',)

    def validate_username(self, value):
        try:
            get_user_model().objects.get(email=value)
            return value
        except ObjectDoesNotExist:
            raise ValidationError("User with this email does not exists")



from django.http import HttpResponse, JsonResponse
from django.urls import reverse
from wagtailcache.cache import cache_page

from geomanager.errors import MissingTileError
from geomanager.models import MBTSource
from geomanager.utils.tile_gl import center_from_bounds, open_mbtiles

DEFAULT_ZOOM = 13
DEFAULT_MINZOOM = 7
DEFAULT_MAXZOOM = 15
WORLD_BOUNDS = [-180, -85.05112877980659, 180, 85.0511287798066]


@cache_page
def tile_gl(request, source_slug, z, x, y):
    source = MBTSource.objects.get(slug=source_slug)
    with (open_mbtiles(source.file.path)) as mbtiles:
        try:
            data = mbtiles.tile(z, x, y)
            response = HttpResponse(
                content=data,
                status=200,
            )
            response["Content-Type"] = "application/x-protobuf"
            response["Content-Encoding"] = "gzip"

            return response

        except MissingTileError:
            return HttpResponse(
                status=204,
            )


@cache_page
def tile_json_gl(request, source_slug):
    source = MBTSource.objects.get(slug=source_slug)
    with open_mbtiles(source.file.path) as mbtiles:
        metadata = mbtiles.metadata()

        # Load valid tilejson keys from the mbtiles metadata
        valid_tilejson_keys = (
            # MUST
            "name",
            "format",
            # SHOULD
            "bounds",
            "center",
            "minzoom",
            "maxzoom",
            # MAY
            "attribution",
            "description",
            "type",
            "version",
            # UNSPECIFIED
            "scheme",
        )
        spec = {key: metadata[key] for key in valid_tilejson_keys if key in metadata}

        if spec["format"] == "pbf":
            spec["vector_layers"] = metadata["json"]["vector_layers"]
        else:
            raise NotImplementedError(
                f"Only mbtiles in pbf format are supported. Found {spec['format']}"
            )

        # Optional fields
        spec["scheme"] = spec.get("scheme", "xyz")
        spec["bounds"] = spec.get("bounds", WORLD_BOUNDS)
        spec["minzoom"] = spec.get("minzoom", DEFAULT_MINZOOM)
        spec["maxzoom"] = spec.get("maxzoom", DEFAULT_MINZOOM)
        spec["center"] = spec.get(
            "center", center_from_bounds(spec["bounds"], DEFAULT_ZOOM)
        )

        # Tile
        tile_url = request.build_absolute_uri(reverse("tile_gl", args=(source.slug, 0, 0, 0)))
        tile_url = tile_url.replace("/0/0/0.pbf", r"/{z}/{x}/{y}.pbf")
        spec["tiles"] = [tile_url]

        # Version
        spec["tilejson"] = "3.0.0"

        return JsonResponse(spec)


@cache_page
def style_json_gl(request, source_slug):
    source = MBTSource.objects.get(slug=source_slug)
    tilejson_url = request.build_absolute_uri(reverse("tile_json_gl", args=[source.slug]))

    style_config = source.json_style

    style_config["id"] = source.pk
    style_config["name"] = source.name
    style_config["glyphs"] = "https://fonts.openmaptiles.org/{fontstack}/{range}.pbf"
    style_config["sources"] = {
        "openmaptiles": {
            "type": "vector", "url": tilejson_url
        }
    }

    return JsonResponse(style_config)
